'''
@Author: YunsangJeong
@Company: Digger
@Digo - Python Logger Module
'''
import requests
import json
import sys
import random
import string
import atexit
import base64

from collections import OrderedDict

from Digo import Project
from Digo import Error
from Digo import Network
from Digo import Utils
from Digo import Singletone

from Digo import ExceptionHandler


class Core():


    DIGO_WEB = 'http://digo.ai:7777/digo-server'


    # def __init__(self, api, project_name):
    #     super().__init__()
    #     self.__user = Project.User(api_key=api)
    #     self.__experiment = Project.Experiment()
    #     self.__project = Project.Project(project_name=project_name)
    #     self.__log: OrderedDict = OrderedDict()
    #     self.__project.name = project_name
    #     self.__step: int = 0
    #     __instance = self
    #     self.login()

    def __init__(self, api, workspace_name, project_name):
        super().__init__()
        self.__user = Project.User(api_key=api)
        self.__experiment = Project.Experiment()
        self.__project = Project.Project(
            project_name=project_name, workspace=workspace_name)
        self.__log: OrderedDict = OrderedDict()
        self.__step: int = 0
        self.login()
        atexit.register(ExceptionHandler.exit_process, self.stateUpdate)

    def send_log(self, log):
        keys = [key for key in log]
        for key in keys:
            value = log[key]
            if type(value) is float:
                log[key] = round(value, 3)

        log["step"] = self.__step
        self.__step += 1
        log_json = json.dumps(log)

        print(log)
        Network.updateLog(api_key=self.__user.api_key, account_id=self.__user.id, project_id=self.__project.id,
                          experiment_id=self.__experiment.id,  log=log_json)

    def printLog(self, log):
        Network.updatePrintLog(api_key=self.__user.api_key, account_id=self.__user.id, project_id=self.__project.id,
                               experiment_id=self.__experiment.id, log=log)

    def send_hyper_parameter(self, hyper_parameter):
        Network.updateHyperParameter(api_key=self.__user.api_key, account_id=self.__user.id, project_id=self.__project.id,
                                     experiment_id=self.__experiment.id, hyper_parameter=hyper_parameter)

    def login(self):
        json_data = Network.login(
            api_key=self.__user.api_key, workspace_name=self.__project.workspace)

        if 'code' not in json_data:
            raise Error.APIInvalidError
            return

        if json_data['code'] == 0:
            self.__user.id = json_data['account_id']
            self.__user.sas_token = json_data['sas_token']
            self.__user.setContainer(json_data['cloud_container_name'])
            print('Login Complete, Welcome {0}'.format(self.__user.id))
            self.create_experiment()
            self.uploadSourceCode()
            self.uploadRequirement()

            
        else:
            # print(json_data)
            raise Error.APIInvalidError

    def create_experiment(self):
        experiment_name: str = Utils.create_experiment_name()

        conn: requests.Response = None

        if self.__project.workspace is None:
            conn = requests.get(self.DIGO_WEB + "/CreateExperiment", params={
                                'api_key': self.__user.api_key, 'account_id': self.__user.id, 'project_name': self.__project.name, 'experiment_name': experiment_name})
        else:
            params = {'api_key': self.__user.api_key, 'account_id': self.__user.id, 'project_name': self.__project.name,
                      'experiment_name': experiment_name, 'workspace_name': self.__project.workspace}
            conn = requests.get(
                self.DIGO_WEB + "/CreateExperiment", params=params)
            # print(params)

        json_data = conn.json()
        # print (json_data)

        if json_data['code'] == 0:
            self.__experiment.id = json_data['experiment_id']
            self.__experiment.name = experiment_name
            self.__project.id = json_data['project_id']
            print("Created Experiment : {0}".format(experiment_name))
        else:
            
            raise Error.FailExperiment

    def uploadSourceCode(self):
        Network.uploadSourceCode(experiment_name=self.__experiment.name,
                                 experiment_id=self.__experiment.id,
                                 project_id=self.__project.id,
                                 sas_token=self.__user.sas_token,
                                 container_name=self.__user.container,
                                 project_name=self.__project.name, api_key=self.__user.api_key, account_id=self.__user.id)

    def uploadRequirement(self):
        Network.uploadRequirementsFile(api_key=self.__user.api_key, account_id=self.__user.id, project_id=self.__project.id,
                                       experiment_id=self.__experiment.id, requirements=Utils.getRequirements())

    def imageLog(self, log):
        Network.uploadImageLog(api_key=self.__user.api_key, account_id=self.__user.id, sas_token=self.__user.sas_token, project_id=self.__project.id,
                               experiment_id=self.__experiment.id, container_name=self.__user.container, project_name=self.__project.name, experiment_name=self.__experiment.name, imageLog=log)

    def uploadModel(self, model_path):
        Network.uploadModelFile(experiment_name=self.__experiment.name,
                                 experiment_id=self.__experiment.id,
                                 project_id=self.__project.id,
                                 sas_token=self.__user.sas_token,
                                 container_name=self.__user.container,
                                 project_name=self.__project.name, api_key=self.__user.api_key, account_id=self.__user.id
                                 ,model_path=model_path)

    def stateUpdate(self, state):
        # state 0 : running, 1 : finished, 2 crashed
        Network.updateExperimentState(api_key=self.__user.api_key, experiment_id=self.__experiment.id, project_id=self.__project.id, account_id=self.__user.id, state=state)

    def upload_test(self, file):
        Network.upload_file_to_main(api_key=self.__user.api_key, account_id=self.__user.id, file=file, file_type = Utils.FILETYPE.IMAGE.name)


import sys, traceback
import atexit
from Digo import DigoCore

isCrash = False

def check_crash(exc_type, exc_value, exc_traceback):
	global isCrash
	isCrash = True
	traceback.print_exception(exc_type, exc_value, exc_traceback)

def exit_process(stateUpdate):
	if isCrash is True:
		stateUpdate(2)
		print("Digo Exit(1)")
	else:
		stateUpdate(1)
		print("Digo Exit(0)")



sys.excepthook = check_crash



class Singletone():
    __instance = None

    @classmethod
    def __getInstance(cls):
        return cls.__instance

    @classmethod
    def instance(cls, *args, **kargs):
        cls.__instance = cls(*args, **kargs)
        cls.instance = cls.__getInstance
        return cls.__instance



import requests
from Digo import SystemMetrics
import sys
import os
import json
import random
import numpy as np
from PIL import Image
import string
from Digo.AzureCloud import AzureCloudService
from collections import OrderedDict

from Digo.Utils import FILETYPE

DIGO_WEB = 'http://digo.ai:7777/digo-server'
# DIGO_WEB = 'http://localhost:8080'

azure_cloud_service = AzureCloudService()
imageLogStep = 0
modelFileStep = 0


def login(api_key, workspace_name):
    conn = requests.get(DIGO_WEB + "/SignInUseAPI",
                        params={'api_key': api_key, 'workspace_name': workspace_name})

    return conn.json()


def updateLog(api_key , account_id, project_id, experiment_id, log):
    conn = requests.post(DIGO_WEB + "/UpdateLog", params={
        'api_key': api_key, 'account_id': account_id, 'project_id':project_id, 'experiment_id': experiment_id, 'log': log, 'system_metrics': SystemMetrics.getSystemMetrics()})


def updatePrintLog(api_key, account_id, project_id, experiment_id, log):
    conn = requests.post(DIGO_WEB + "/UpdatePrintLog", params={
        'api_key': api_key, 'account_id': account_id, 'project_id':project_id, 'experiment_id': experiment_id, 'print_log': log})


def updateHyperParameter(api_key, account_id, project_id, experiment_id, hyper_parameter):
    conn = requests.post(DIGO_WEB + "/UpdateHyperParameter", params={
        'api_key': api_key, 'account_id': account_id,'project_id':project_id, 'experiment_id': experiment_id, 'hyper_parameter': hyper_parameter})

def updateExperimentState(api_key, account_id,experiment_id, project_id, state):
    conn = requests.post(DIGO_WEB + "/UpdateExperimentState", params={
        'api_key': api_key, 'account_id': account_id, 'experiment_id': experiment_id,'project_id':project_id, 'experiment_state':state})


def upload_file_to_main(file, api_key, account_id, file_type):
    conn = requests.post(DIGO_WEB + "/UploadFileForModule", params={ 'api_key': api_key, 'account_id': account_id, 'file_type':file_type}, files={"file":file})
    return conn.json()["file_index"]


def uploadModelFile(api_key, account_id,project_id, experiment_name, sas_token, container_name, project_name, experiment_id, model_path):
    global modelFileStep
    current_dir = os.getcwd()
    origin_file_name = model_path.split('/')[-1]
    file_name = str(modelFileStep) + "_" + origin_file_name
    os.rename

    if model_path.startswith('/'):
        full_path = current_dir + model_path
    else:
        full_path = current_dir + os.sep + model_path

    convert_full_path = full_path.split(origin_file_name)[0] + file_name

    os.rename(full_path, convert_full_path)

    # print(convert_full_path)

    blob_id = upload_file_to_main(file=open(convert_full_path, 'rb'), account_id= account_id, api_key=api_key, file_type=FILETYPE.MODEL.name)

    model_blob = OrderedDict()
    model_blob[file_name] = blob_id
    os.rename(convert_full_path, full_path)
    conn = requests.post(DIGO_WEB + "/ModelFileAzureUpdateComplete", params={
        'api_key': api_key, 'account_id': account_id, 'project_id':project_id, 'experiment_id': experiment_id, 'model_blob': json.dumps(model_blob)})
    modelFileStep +=1

def uploadSourceCode(api_key, account_id, project_id, experiment_name, sas_token, container_name, project_name, experiment_id):

    local_file_name = sys.argv[0]
    directory, file_name = os.path.split(local_file_name)
    current_dir = os.getcwd()
    full_path = current_dir + os.sep + file_name

    blob_id = upload_file_to_main(file=open(local_file_name, 'rb'), api_key=api_key, account_id=account_id, file_type=FILETYPE.SOURCE.name)

    conn = requests.post(DIGO_WEB + "/SourceCodeAzureUpdateComplete", params={
        'api_key': api_key, 'account_id': account_id, 'project_id':project_id, 'experiment_id': experiment_id, 'azure_blob_id': blob_id})


def uploadRequirementsFile(api_key, account_id, project_id, experiment_id, requirements):
    conn = requests.post(DIGO_WEB + "/UploadRequirements", params={
        'api_key': api_key, 'account_id': account_id, 'project_id':project_id, 'experiment_id': experiment_id, 'requirements': requirements})


def uploadImageLog(api_key, account_id, sas_token, project_id, experiment_id, container_name, project_name, experiment_name, imageLog):
    global imageLogStep

    local_file_name = sys.argv[0]
    directory, file_name = os.path.split(local_file_name)
    current_dir = os.getcwd()
    full_path = current_dir + os.sep + local_file_name
    # print(full_path)

    result_image_log = OrderedDict()

    if os.path.isdir(current_dir+ os.sep + 'temp') is False :
        os.mkdir(current_dir+ os.sep + 'temp')


    for key in imageLog:
        if type(imageLog[key]) is np.ndarray:
            image_data = Image.fromarray(imageLog[key])
        else:
            image_data = imageLog[key]
        image_name: str = key + "_" + str(imageLogStep)
        image_data.save(current_dir+ os.sep + 'temp' + os.sep + image_name + '.png', 'PNG')
        full_path = current_dir + os.sep + 'temp' + os.sep + image_name +'.png'
        result_image_log[key] = upload_file_to_main(file=open(full_path, 'rb'), api_key=api_key, account_id=account_id, file_type=FILETYPE.IMAGE.name)
        # azure_cloud_service.upload_image(full_path, image_name+'.png', sas_token, container_name, project_name, experiment_name, experiment_id)
        os.remove(current_dir + os.sep + 'temp' + os.sep + image_name + '.png')

    conn = requests.post(DIGO_WEB + "/ImageAzureUpdateComplete", params={
        'api_key': api_key, 'account_id': account_id, 'project_id':project_id, 'experiment_id': experiment_id, 'image_blob_id': json.dumps(result_image_log)})

    imageLogStep += 1



from azure.storage.blob import BlobServiceClient
from azure.storage.blob import ContainerClient

import os
import requests
import random
import string


# server parameter
DIGO_WEB = 'http://digo.ai:7777/digo-server'
# DIGO_WEB = 'http://localhost:8080/'


class AzureCloudService:

    def upload_source_file(self,
                           full_path_to_file,
                           file_name,
                           sas_token,
                           container_name,
                           project_name,
                           experiment_name,
                           experiment_id):

        return self._save_blob_to_cloud(full_path_to_file,
                                        file_name,
                                        'Source Code',
                                        sas_token,
                                        container_name,
                                        project_name,
                                        experiment_name,
                                        experiment_id)

    def upload_model_file(self,
                          full_path_to_file,
                          file_name,
                          sas_token,
                          container_name,
                          project_name,
                          experiment_name,
                          experiment_id):

        return self._save_blob_to_cloud(full_path_to_file,
                                        file_name,
                                        'Model',
                                        sas_token,
                                        container_name,
                                        project_name,
                                        experiment_name,
                                        experiment_id)

    def upload_image(self,
                     full_path_to_file,
                     file_name,
                     sas_token,
                     container_name,
                     project_name,
                     experiment_name,
                     experiment_id):

        return self._save_blob_to_cloud(full_path_to_file,
                                        file_name,
                                        'Image',
                                        sas_token,
                                        container_name,
                                        project_name,
                                        experiment_name,
                                        experiment_id)

    def _upload_to_cloud(self,
                         data,
                         file_name,
                         file_type,
                         project_name,
                         experiment_name,
                         blob_service_client: BlobServiceClient,
                         container_name: str):

        random_pool = string.ascii_letters + string.digits
        random_len = 8
        random_str = ''
        for _ in range(random_len):
            random_str += random.choice(random_pool)

        file_dir, temp = os.path.split(data.name)
        save_name = project_name + os.path.sep + experiment_name + os.path.sep

        if file_type == 'Source Code':
            save_name = save_name + 'source' + os.sep + file_name
        elif file_type == 'Image':
            save_name = save_name + 'images' + os.sep + file_name
        elif file_type == 'Model':
            save_name = save_name + 'model' + os.sep + file_name
        else:
            raise Exception('Invalid File Type!!!')

        container_name = container_name.strip()

        # print('Uploading... ' + save_name)
        blob_client = blob_service_client.get_blob_client(blob=save_name)
        blob_client.upload_blob(data, overwrite=True)
        # print('Upload Succeed!!')

    def _build_connection_string(self,
                                 account_name,
                                 account_key):
        connection_string = 'DefaultEndpointsProtocol=https' + ';'\
                            'AccountName=' + account_name + ';' \
                            'AccountKey=' + account_key + ';' \
                            'EndpointSuffix=core.windows.net'

        return connection_string

    def _save_blob_to_cloud(self,
                            full_path_to_file,
                            file_name,
                            file_type,
                            sas_token,
                            container_name,
                            project_name,
                            experiment_name,
                            experiment_id):

        with open(full_path_to_file, 'rb') as file:
            # later: use 'isinstance' to check instance.
            conn: requests.Response = None

            file_size = os.path.getsize(file.name)

            if file_size == 0:
                raise Exception('No File Detected!!!')

            params = {'container_name': container_name,
                      'project_name': project_name,
                      'experiment_name': experiment_name,
                      'experiment_id': experiment_id,
                      'file_name': file_name,
                      'file_size': file_size,
                      'file_type': file_type}

            conn = requests.get(
                DIGO_WEB + "/CheckAzureCloud", params=params)

            json_data = conn.json()

            # print(conn.url)

            # Response from server
            if json_data['code'] == 0:
                blob_service_client = ContainerClient.from_container_url(
                    container_url="https://digocloud.blob.core.windows.net/"+container_name,
                    credential=sas_token
                )

                # BlobServiceClient.from_connection_string(connection_string)

                self._upload_to_cloud(file,
                                      file_name,
                                      file_type,
                                      project_name,
                                      experiment_name,
                                      blob_service_client,
                                      container_name,)

                return json_data['azure_blob_id']

            elif json_data['code'] == -200:
                raise Exception('Not Enough Storage Left!!!')



class APIInvalidError(Exception):
    def __init__(self):
        super().__init__('API key is invalid, Please check the API key again on http://digo.ai')

class NotFoundCore(Exception):
    def __init__(self):
        super().__init__('Digo Core is Not Found')

class FailExperiment(Exception):
    def __init__(self):
        super().__init__('Fail Experiment')

class InvalidType(Exception):
    def __init__(self):
        super().__init__('Invalid Input Type')


class User:
    def __init__(self, api_key, container = None, sasToken = None):
        super().__init__()
        self.id:str
        self.api_key:str = api_key
        self.container:str = container
        self.sas_token: str = sasToken

    def setContainer(self, container):
        self.container:str = container
    
    def setAzureKey(self, id, key):
        self.azure_id:str = id
        self.azure_key:str = key

class Project:
    def __init__(self, project_name, workspace = None):
        super().__init__()
        self.workspace:str = workspace
        self.id:str
        self.name:str = project_name

class Experiment:
    def __init__(self):
        super().__init__()
        self.name:str
        self.id:str
        


import psutil
import nvgpu

from collections import OrderedDict
import json

def getSystemMetrics():

    systemMetrics = OrderedDict()
    systemMetrics["cpu_usage"] = psutil.cpu_percent()
    systemMetrics["memory_usage"] = psutil.virtual_memory().percent

    try : 
        systemMetrics["gpu_usage"] = round(nvgpu.gpu_info()[0]["mem_used_percent"],1)
    except FileNotFoundError:
        pass

    return json.dumps(systemMetrics)


from setuptools import setup, find_packages

setup(
    name='Digo',
    version='0.0.10',
    author='DiggerWorks',
    author_email='support@digger.works',
    packages=find_packages(),
    long_description=open('README.md').read(),
    keywords=['digo', 'ml'],
    install_requires=['azure-core', 'azure-storage-blob', 'Pillow', 'numpy', 'requests', 'psutil', 'nvgpu', 'argparse'],
    zip_safe=False
)


# __init__.py
# Copyright (C) 2020 (support@digger.works) and digger


__author__ = 'DiggerWorks'
__email__ = 'support@digger.works'
__version__ = '0.0.10'

from Digo import DigoCore
from Digo import Error
from Digo import ExceptionHandler
import json
import argparse

core = None


def init(api_key: str, workspace_name: str, project_name: str):
    global core
    core = DigoCore.Core(
        api=api_key, workspace_name=workspace_name, project_name=project_name)


def log(log):
    if core is None:
        raise Error.NotFoundCore
    core.send_log(log)


def imageLog(log):
    if type(log) == dict:
        core.imageLog(log)
    else:
        raise Error.InvalidType

def upload_test(file):
    core.upload_test(file)

def uploadModel(model_path):
    core.uploadModel(model_path)

def setParameter(hyper_parameter):
    if core is None:
        raise Error.NotFoundCore

    if type(hyper_parameter) == argparse.ArgumentParser:
        parameter = vars(hyper_parameter.parse_args())
        print(json.dumps(parameter))
        core.send_hyper_parameter(json.dumps(parameter))
    elif type(hyper_parameter) == dict:
        core.send_hyper_parameter(json.dumps(parameter))
    else:
        raise Error.InvalidType


def printf(log):
    if core is None:
        raise Error.NotFoundCore

    print(log)
    core.printLog(log)



import subprocess
import os
import random
from enum import Enum

class FILETYPE(Enum):
    IMAGE=0
    SOUND=1
    MOVIE=2
    TEXT=3
    JSON=4
    SOURCE=5
    MODEL=6

def getRequirements():
    proc = subprocess.Popen(
        ['pip', 'freeze'],
        stdout=subprocess.PIPE
    )

    out, err = proc.communicate()
    requirements = out.decode('utf-8')
    return requirements

def create_experiment_name():
    adjectives = ['brilliant', 'adorable', 'shy', 'beautiful', 'popular', 'huge', 'curious', 'confident', 'assured', 'eager', 'energetic', 'loyal', 'nervous', 'careful', 'relaxed', 'calm', 'component', 'capable', 'talented', 'kind', 'wise', 'bright', 'busy', 'cool', 'cute', 'fair', 'fine', 'fresh', 'funny', 'grand', 'great', 'glad', 'humorous', 'lucky', 'nice', 'perfect', 'polite', 'proud', 'powerful', 'pretty', 'pure', 'quick', 'quiet', 'real', 'rich', 'right', 'royal', 'social', 'soft', 'safe', 'serious', 'simple', 'still', 'strong', 'thin', 'tiny', 'tired', 'virtual', 'weak', 'wide', 'warm', 'wild','young', 'lazy', 'lonely', 'large', 'angry', 'asleep', 'brave', 'active']
    objectives = ['crocodile', 'ant' ,'eater', 'bat', 'beaver', 'buffalo', 'camel', 'chameleon', 'cheetah', 'squirrel', 'chinchilla', 'cormorant', 'coyote', 'crow', 'dinosaur', 'dolphin', 'duck', 'elephant', 'fox', 'weasel', 'frog', 'giraffe', 'bear', 'hedgehog', 'hippo', 'hyena', 'iguana', 'koala', 'leopard', 'liger', 'llama', 'monkey', 'whale', 'cat', 'orangutan', 'otter', 'panda', 'penguin', 'platypus', 'raccoon', 'rhinoceros', 'sheep', 'skunk', 'turtle', 'wolf', 'badger', 'gorilla', 'snail', 'dog', 'pig', 'chicken']
    
    experiment_name : str = random.choice(adjectives) + "-" + random.choice(objectives) + "-"
    numbers = "0123456789"
    for i in range(0, 4):
            experiment_name += random.choice(numbers)

    return experiment_name